import pathlib
from typing import Union

import attr

import paralleldots
import requests
from aylienapiclient import textapi

from srec_nlp.common import _ensure_utf, Query, Score, _parallel
from copy import deepcopy
from requests import HTTPError


class NLP:
    @_parallel
    @_ensure_utf
    def sentiment(self, text) -> Query:
        """
        Sample class.
        :param text: Union[IOBase, str, Path, bytes] a file, pathtofile, bytes or string with the text.
                    Function guarantees the value will be converted to text
        :return: Query
        """
        raise NotImplemented()

    def copy(self):
        return deepcopy(self)


@attr.s
class ParallelDots(NLP):
    """
       with ParallelDots(api_key) as nlp:
               nlp.sentiment(text)
               nlp.sentiment(bytes)
               nlp.sentiment(open(path/to/text.txt,'r'))
               nlp.sentiment(pathlib.Path('path/to/text.txt'))

       """

    api_key: str = attr.ib()

    @api_key.validator
    def setkey(self, _, value):
        paralleldots.set_api_key(value)

    def __enter__(self):
        return self

    @_parallel
    @_ensure_utf
    def sentiment(self, text) -> Query:
        """
        Produces a sentiment score for query text using ParallelDots API
        :param text: Union[IOBase, str, Path, bytes] a file, pathtofile, bytes or string with the text.
                    Function guarantees the value will be converted to text
        :return: Query
        """
        try:
            response = paralleldots.sentiment(text)
            return Query(text, Score.fromtriplet(**response["sentiment"]))
        except (HTTPError, KeyError) as _:
            raise HTTPError()

    def __exit__(self, *exc):
        pass


class TextProcessing(NLP):
    """
    with TextProcessing() as nlp:
            nlp.sentiment(text)
            nlp.sentiment(bytes)
            nlp.sentiment(open(path/to/text.txt,'r'))
            nlp.sentiment(pathlib.Path('path/to/text.txt'))

        # TextProcessing is still usable at this point.
        # So reusing it is still valid

        nlp.sentiment(text)
        nlp.sentiment(bytes)
        nlp.sentiment(open(path/to/text.txt,'r'))
        nlp.sentiment(pathlib.Path('path/to/text.txt'))
    """

    _url = "http://text-processing.com/api/sentiment/"

    def __enter__(self):
        return self

    @_parallel
    @_ensure_utf
    def sentiment(self, text) -> Query:
        """
        Produces a sentiment score for query text using TextProcessing API
        :param text: Union[IOBase, str, Path, bytes] a file, pathtofile, bytes or string with the text.
                    Function guarantees the value will be converted to text
        :return: Query
        """
        data = dict(text=text)
        result = requests.post(self._url, data=data).json()
        result = result["probability"]
        result["positive"] = result["pos"]
        result["negative"] = result["neg"]
        del result["pos"]
        del result["neg"]
        try:
            return Query(content=text, score=Score.fromtriplet(**result))
        except (HTTPError, KeyError) as _:
            raise HTTPError()

    def __exit__(self, *exc):
        pass


try:
    import fastText


    class FastText(NLP):
        """
        model: str/pathlib.Path  the path to the model.

        # Usage example:

        with FastText(path/to/model.bin) as nlp:
            nlp.sentiment(text)
            nlp.sentiment(bytes)
            nlp.sentiment(open(path/to/text.txt,'r'))
            nlp.sentiment(pathlib.Path('path/to/text.txt'))

        # at this point, nlp.sentiment WILL RAISE AN ERROR because the model has been
        # dealloc'd, to avoid this, use the following:

        nlp = FastText(path/to/model.bin)
        nlp.sentiment(text)
        nlp.sentiment(bytes)
        nlp.sentiment(open(path/to/text.txt,'r'))
        nlp.sentiment(pathlib.Path('path/to/text.txt'))

        """

        def __init__(self, path: Union[str, pathlib.Path]):
            self.model = FastText._load_model(path)
            self.path = path

        def __enter__(self):
            return self

        def __exit__(self, *vargs, **kwargs):
            del self.model

        @_parallel
        @_ensure_utf
        def sentiment(self, text) -> Query:
            """
            Produces a sentiment score for query text using FastText API
            Note that FastText needs to be installed for this to work.
            :param text: Union[IOBase, str, Path, bytes] a file, pathtofile, bytes or string with the text.
                        Function guarantees the value will be converted to text
            :return: Query
            """
            labels, scores = self.model.predict(text, 1)
            label = labels[0]
            score = scores[0]
            if label == "__label__1":
                score = 1 - score
            score = Score.fromsingle(score)
            return Query(text, score)

        @staticmethod
        def _load_model(path: Union[str, pathlib.Path]):
            p = pathlib.Path(path).expanduser()
            if not p.is_file():
                raise ValueError(f"{p} is not a file.")
            try:
                return fastText.load_model(str(p))
            except Exception as e:
                print(e)
                raise ValueError(f"{p} is not a valid FastText model")

        def copy(self):
            return FastText(self.path)


except:
    pass


class Aylien(NLP):
    def __init__(self, appid, appkey):
        """
        :param appid: string Aylien application id
        :param appkey: string Aylien application key

        with Aylien(id, key) as nlp:
            nlp.sentiment(text)
            nlp.sentiment(bytes)
            nlp.sentiment(open(path/to/text.txt,'r'))
            nlp.sentiment(pathlib.Path('path/to/text.txt'))

        # Aylien is still usable at this point.
        # So reusing it is still valid

        nlp.sentiment(text)
        nlp.sentiment(bytes)
        nlp.sentiment(open(path/to/text.txt,'r'))
        nlp.sentiment(pathlib.Path('path/to/text.txt'))
        """
        self.client = textapi.Client(applicationId=appid, applicationKey=appkey)

    def __enter__(self):
        return self

    @_parallel
    @_ensure_utf
    def sentiment(self, text: str) -> Query:
        """
        Produces a sentiment score for query text using Aylien API
        :param text: Union[IOBase, str, Path, bytes] a file, pathtofile, bytes or string with the text.
                    Function guarantees the value will be converted to text
        :return: Query
        """
        sentiment = self.client.Sentiment(dict(text=text))
        try:
            confidence = sentiment["polarity_confidence"]
            klass = sentiment["polarity"]
            remaining = 1 - confidence
            vals = dict(positive=remaining, negative=remaining)
            vals[klass] = confidence
            return Query(content=text, score=Score(**vals))
        except (HTTPError, KeyError) as _:
            raise HTTPError()

    def __exit__(self, *exc):
        pass



import functools
import itertools
from io import IOBase
from pathlib import Path
from typing import List, Dict, Callable, Union
import copy
import attr
import asyncio
import trio
from requests import HTTPError
from trio import MultiError


@attr.s(frozen=True)
class Score:
    positive: float = attr.ib(factory=float)
    negative: float = attr.ib(factory=float)
    classification: str = ""

    def __attrs_post_init__(self):
        classes = ("positive", "negative")
        filt = [getattr(self, c) for c in classes]
        mx = max(filt)
        object.__setattr__(
            self,
            "classification",
            next(itertools.compress(classes, map(mx.__eq__, filt))),
        )

    @classmethod
    def fromsingle(cls, value):
        """
        :param value: single value score
        :return:
        """
        negative = 1 - value
        positive = value

        return Score(positive=positive, negative=negative)

    @classmethod
    def fromtriplet(cls, positive, neutral, negative):
        """
        :param positive:  Confidence
        :param neutral: Confidence
        :param negative: Confidence
        :return: Score
        """
        positive += neutral / 2
        negative += neutral / 2
        return Score(positive=positive, negative=negative)


@attr.s(frozen=True)
class Query:
    content: str = attr.ib(factory=str)
    score: Score = attr.ib(default=Score())


def _ensure_utf(
        func: Callable[[str], Query]
) -> Callable[[Union[IOBase, str, Path, bytes]], Query]:
    """

    :param func: Query function to wrap
    :return: Returns a wrapper for the provided function.
    """

    @functools.singledispatch
    def _text_is_utf(text: str):
        return text.replace("\n", "")

    @_text_is_utf.register(bytes)
    def _(text: bytes):
        return text.decode("utf-8").replace("\n", "")

    @_text_is_utf.register(Path)
    def _(path: Path):
        text = path.read_text("utf-8").replace("\n", "")
        return text

    @_text_is_utf.register(IOBase)
    def _(fp: IOBase):
        text = "".join(map(str, fp.readlines())).replace("\n", "")
        return text

    @functools.wraps(func)
    def wrapper(self, text):
        text = _text_is_utf(text).replace("\n", "")
        return func(self, text)

    return wrapper


def _parallel(func):
    async def listd(self, query):
        def catch_httperr(exc):
            if isinstance(exc, HTTPError):
                print(exc)
                raise HTTPError()
            else:
                raise MultiError([exc])

        async def list_internal(s, q, callback):
            callback(func(s, q))

        results = []
        with trio.MultiError.catch(catch_httperr):
            async with trio.open_nursery() as nursery:
                for q in query:
                    nursery.start_soon(list_internal, self, q, results.append)

        return results

    @functools.wraps(func)
    def internal(self, query):
        if isinstance(query, list):
            try:
                return trio.run(listd, self, query)
            except trio.MultiError as err:
                exceptions = err.exceptions()

        return func(self, query)

    return internal



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['srec_nlp']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.9,<0.30.0',
 'attrs>=19.1,<20.0',
 'aylien-apiclient>=0.7.0,<0.8.0',
 'google-cloud-language>=1.2,<2.0',
 'google-cloud>=0.34.0,<0.35.0',
 'ibm-watson>=3.0,<4.0',
 'numpy>=1.16,<2.0',
 'paralleldots>=3.2,<4.0',
 'pybind11>=2.2,<3.0',
 'requests>=2.22,<3.0',
 'sklearn>=0.0.0,<0.0.1',
 'trio>=0.11.0,<0.12.0']

setup_kwargs = {
    'name': 'srec-nlp',
    'version': '0.2.6',
    'description': '',
    'long_description': None,
    'author': 'Stelios Tymvios',
    'author_email': 'stelios.tymvios@icloud.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
